from dataclasses import dataclass, field

@dataclass(frozen=True)
class Component:
    type : str
    id : str = field(default='0')
    nodes : tuple[str, ...] = field(default=('0',))
    value: dict[str, float | str] = field(default_factory=dict)

def resistor(id: str, nodes: tuple[str, str], R: float) -> Component:
    if R < 0:
        raise ValueError('R must be greater than zero.')
    return Component(
        type='resistor',
        id=id,
        value={'R': R},
        nodes=nodes
        )

def conductance(id: str, nodes: tuple[str, str], G: float) -> Component:
    if G < 0:
        raise ValueError('G must be greater than zero.')
    return Component(
        type='conductance',
        id=id,
        value={'G': G},
        nodes=nodes
        )

def capacitor(id: str, nodes: tuple[str, str], C: float) -> Component:
    if C < 0:
        raise ValueError('C must be greater than zero.')
    return Component(
        type='capacitor',
        id=id,
        value={'C': C},
        nodes=nodes
    )

def inductance(id: str, nodes: tuple[str, str], L: float) -> Component:
    if L < 0:
        raise ValueError('L must be greater than zero.')
    return Component(
        type='inductance',
        id=id,
        value={'L': L},
        nodes=nodes
    )

def impedance(id: str, nodes: tuple[str, str], Z: complex) -> Component:
    return Component(
        type='impedance',
        id=id,
        value={'R': Z.real, 'X': Z.imag},
        nodes=nodes
        )

def admittance(id: str, nodes: tuple[str, str], Y: complex) -> Component:
    return Component(
        type='admittance',
        id=id,
        value={'G': Y.real, 'B': Y.imag},
        nodes=nodes
        )

def dc_voltage_source(id: str, nodes: tuple[str, str], V: float, R: float = 0) -> Component:
    if R < 0:
        raise ValueError('R must be greater than zero.')
    return Component(
        type='dc_voltage_source',
        id=id,
        value={'V': V, 'R': R, 'w': 0, 'phi': 0},
        nodes=nodes
        )

def ac_voltage_source(id: str, nodes: tuple[str, str], V: float, R: float = 0, w: float = 0, phi: float = 0) -> Component:
    if R < 0:
        raise ValueError('R must be greater than zero.')
    if w < 0:
        raise ValueError('w must be greater than zero.')
    return Component(
        type='ac_voltage_source',
        id=id,
        value={'V': V, 'R': R, 'w': w, 'phi': phi},
        nodes=nodes
        )

def complex_voltage_source(id: str, nodes: tuple[str, str], V: complex, Z: complex = 0) -> Component:
    return Component(
        type='complex_voltage_source',
        id=id,
        value={'V_real': V.real, 'V_imag': V.imag, 'R': Z.real, 'X': Z.imag},
        nodes=nodes
        )

def periodic_voltage_source(id: str, nodes: tuple[str, str], wavetype: str, V: float, w: float, phi: float = 0, R: float = 0) ->Component:
    if R < 0:
        raise ValueError('R must be greater than zero.')
    if w < 0:
        raise ValueError('w must be greater than zero.')
    return Component(
        type='periodic_voltage_source',
        id=id,
        value={'wavetype': wavetype,
               'V': V,
               'w': w,
               'phi': phi,
               'R': R},
        nodes=nodes
        )

def dc_current_source(id: str, nodes: tuple[str, str], I: float, G: float = 0) -> Component:
    if G < 0:
        raise ValueError('G must be greater than zero.')
    return Component(
        type='dc_current_source',
        id=id,
        value={'I': I, 'G': G, 'w': 0, 'phi': 0},
        nodes=nodes
        )

def ac_current_source(id: str, nodes: tuple[str, str], I: float, G: float = 0, w: float = 0, phi: float = 0) -> Component:
    if G < 0:
        raise ValueError('G must be greater than zero.')
    if w < 0:
        raise ValueError('w must be greater than zero.')
    return Component(
        type='ac_current_source',
        id=id,
        value={'I': I, 'G': G, 'w': w, 'phi': phi},
        nodes=nodes
        )

def complex_current_source(id: str, nodes: tuple[str, str], I: complex, Y: complex = 0) -> Component:
    return Component(
        type='complex_current_source',
        id=id,
        value={'I_real': I.real, 'I_imag': I.imag, 'G': Y.real, 'B': Y.imag},
        nodes=nodes
        )

def periodic_current_source(id: str, nodes: tuple[str, str], wavetype: str, I: float, w: float, phi: float, G: float = 0) -> Component:
    return Component(
        type='periodic_current_source',
        id=id,
        value={'wavetype': wavetype,
               'I': I,
               'w': w,
               'phi': phi,
               'G': G},
        nodes=nodes
        )

def lamp(id: str, nodes: tuple[str, str], P: float, V_ref: float) -> Component:
    if P < 0:
        raise ValueError('P must be greater than zero.')
    if V_ref < 0:
        raise ValueError('V_ref must be greater than zero.')
    return Component(
        type='lamp',
        id=id,
        value={'P': P, 'V_ref': V_ref},
        nodes=nodes
        )

def resistive_load(id: str, nodes: tuple[str, str], P: float, V_ref: float) -> Component:
    if P < 0:
        raise ValueError('P must be greater than zero.')
    if V_ref < 0:
        raise ValueError('V_ref must be greater than zero.')
    return Component(
        type='resistive_load',
        id=id,
        value={'P': P, 'V_ref': V_ref},
        nodes=nodes
        )

def short_circuit(id: str, nodes: tuple[str, str]) -> Component:
    return Component(
        type='short_circuit',
        id=id,
        nodes=nodes
    )        

def ground(id: str='gnd', nodes: tuple[str]=('0',)) -> Component:
    return Component(
        type='ground',
        id=id,
        nodes=nodes
    )


from . import layout
from .plot_elements import complex_pointer
import functools
from typing import Callable, Dict, TypedDict
from ..Circuit.solution import ComplexSolution

def plot_pointer_by_id(id:str, origin:str='', scaling:float=1, reverse:bool=False, **kwargs) -> layout.PlotFcn:
    def plot_pointer(fig:layout.Figure, ax:layout.Axes, *, pointer_fcn:Callable[[str], complex]=lambda _: 0+0j, label_fcn:Callable[[str], str]=lambda _: '') -> layout.FigureAxes:
        z0 = 0 if origin == '' else complex(pointer_fcn(origin))
        z1 = complex(pointer_fcn(id))
        z1 *= scaling
        if reverse:
            z1 = -z1
        if 'label' not in kwargs.keys():
            kwargs.update({'label':label_fcn(id)})
        complex_pointer(ax[0], z0, z0+z1, **kwargs)
        return fig, ax
    return plot_pointer

def plot_pointer_by_value(z:complex, origin:complex=0+0j, **kwargs) -> layout.PlotFcn:
    def plot_pointer(fig:layout.Figure, ax:layout.Axes, **_) -> layout.FigureAxes:
        complex_pointer(ax[0], origin, z+origin, **kwargs)
        return fig, ax
    return plot_pointer

class PlotPointerDiagramProperties(TypedDict):
    pointer_fcn: Callable[[str], complex]
    label_fcn: Callable[[str], str]
    xlabel:str
    ylabel:str

def plot_pointers_factory(
        *args,
        type:str='default',
        solution:ComplexSolution,
        pd_lim:tuple[float, float, float, float]=(-1, 1, -1, 1),
        xlabel:str='',
        ylabel:str='') -> layout.PlotFcn:
    
    pointer_diagram_configurations: Dict[str, PlotPointerDiagramProperties] = {
        'default' : {'pointer_fcn' : lambda _: 0j, 'label_fcn' : lambda _: '', 'xlabel' : 'Re→', 'ylabel' : 'Im→'},
        'voltage' : {'pointer_fcn' : solution.get_voltage, 'label_fcn' : lambda id: f'V({id})', 'xlabel' : 'Re{V}→', 'ylabel' : 'Im{V}→'},
        'current' : {'pointer_fcn' : solution.get_current, 'label_fcn' : lambda id: f'I({id})', 'xlabel' : 'Re{I}→', 'ylabel' : 'Im{I}→'},
        'power' : {'pointer_fcn' : solution.get_power, 'label_fcn' : lambda id: f'S({id})', 'xlabel' : 'P→', 'ylabel' : 'Q→'}
    }
    pd_properties = pointer_diagram_configurations.get(type, pointer_diagram_configurations['default'])
    xlabel = xlabel if len(xlabel) > 0 else pd_properties['xlabel']
    ylabel = ylabel if len(ylabel) > 0 else pd_properties['ylabel']

    @layout.legend()
    @layout.grid()
    @layout.nyquist_like_plot(ax_lim=pd_lim, xlabel=xlabel, ylabel=ylabel)
    def plot_pointers(fig:layout.Figure, ax:layout.Axes) -> layout.FigureAxes:
        new_args = tuple(functools.partial(a, pointer_fcn=pd_properties['pointer_fcn'], label_fcn=pd_properties['label_fcn']) for a in args)
        layout.apply_plt_fcn(fig, ax, *new_args)
        return fig, ax
    return plot_pointers

def pointer_diagram(*args, layout_fcn:layout.Layout=layout.figure_default, **kwargs) -> layout.FigureAxes:
    plot_pointers = plot_pointers_factory(*args, type='default', **kwargs)
    return plot_pointers(*layout_fcn())

def voltage_pointer_diagram_analysis(*args, layout_fcn:layout.Layout=layout.figure_default, **kwargs) -> layout.FigureAxes:
    plot_pointers = plot_pointers_factory(*args, type='voltage', **kwargs)
    return plot_pointers(*layout_fcn())

def current_pointer_diagram_analysis(*args, layout_fcn:layout.Layout=layout.figure_default, **kwargs) -> layout.FigureAxes:
    plot_pointers = plot_pointers_factory(*args, type='current', **kwargs)
    return plot_pointers(*layout_fcn())

def power_pointer_diagram_analysis(*args, layout_fcn:layout.Layout=layout.figure_default, **kwargs) -> layout.FigureAxes:
    plot_pointers = plot_pointers_factory(*args, type='power', **kwargs)
    return plot_pointers(*layout_fcn())


from typing import Protocol, Type
from abc import ABC, abstractmethod
from dataclasses import dataclass
from .types import TimeDomainFunction
import numpy as np

class UnknownWavetype(Exception):
    ...

class TransformationError(Exception):
    ...

@dataclass
class PeriodicFunction(Protocol):
    period: float
    amplitude: float
    phase: float
    offset: float = 0
    wavetype: str = ''

    @property
    def time_function(self) -> TimeDomainFunction:
        ...

PeriodicFunctionList = list[Type[PeriodicFunction]]

@dataclass
class HarmonicCoefficients(Protocol):
    amplitude0: float
    phase0: float
    offset0: float

    def amplitude(self, n: int) -> float:
        ...

    def phase(self, n: int) -> float:
        ...

    def a(self, n: int) -> float:
        ...
    
    def b(self, n: int) -> float:
        ...

    def c(self, n: int) -> complex:
        ...

@dataclass
class AbstractHarmonicCoefficients(ABC):
    amplitude0: float = 1
    phase0: float = 0
    offset0: float = 0

    def amplitude(self, n: int) -> float:
        if n < 0:
            return self._amplitude_coefficient(-n)
        return self._amplitude_coefficient(n)

    def phase(self, n: int) -> float:
        if n < 0:
            return -self._phase_coefficient(-n)
        return self._phase_coefficient(n)

    def a(self, n: int) -> float:
        return self.amplitude(n)*np.cos(self.phase(n))
    
    def b(self, n: int) -> float:
        return -self.amplitude(n)*np.sin(self.phase(n))

    def c(self, n: int) -> complex:
        if n < 0:
            return self.amplitude(-n)/2*np.exp(-1j*self.phase(-n))
        return self.amplitude(n)/2*np.exp(1j*self.phase(n))

    @abstractmethod
    def _amplitude_coefficient(self, n: int) -> float:
        ...

    @abstractmethod
    def _phase_coefficient(self, n: int) -> float:
        ...

@dataclass
class ConstantFunction:
    period: float = 0
    amplitude: float = 1
    phase: float = 0
    offset: float = 0
    wavetype: str = 'const'

    @property
    def time_function(self) -> TimeDomainFunction:
        return lambda t: self.amplitude*np.ones(t.shape)

class ConstFunctionHarmonics(AbstractHarmonicCoefficients):
    def _amplitude_coefficient(self, n: int) -> float:
        if n == 0:
            return self.amplitude0
        return 0

    def _phase_coefficient(self, _: int) -> float:
        return 0

@dataclass
class CosFunction:
    period: float
    amplitude: float
    phase: float = 0
    offset: float = 0
    wavetype: str = 'cos'

    @property
    def time_function(self) -> TimeDomainFunction:
        return lambda t: self.amplitude*np.cos(2*np.pi/self.period*t + self.phase) + self.offset

class CosFunctionHarmonics(AbstractHarmonicCoefficients):
    def _amplitude_coefficient(self, n: int) -> float:
        if n == 0:
            return self.offset0
        if n == 1:
            return self.amplitude0
        return 0

    def _phase_coefficient(self, n: int) -> float:
        if n == 1:
            return self.phase0
        return 0

@dataclass
class SinFunction:
    period: float
    amplitude: float
    phase: float = 0
    offset: float = 0
    wavetype: str = 'sin'

    @property
    def time_function(self) -> TimeDomainFunction:
        return lambda t: self.amplitude*np.sin(2*np.pi/self.period*t + self.phase) + self.offset

class SinFunctionHarmonics(AbstractHarmonicCoefficients):
    def _amplitude_coefficient(self, n: int) -> float:
        if n == 0:
            return self.offset0
        if n == 1:
            return self.amplitude0
        return 0

    def _phase_coefficient(self, n: int) -> float:
        if n == 1:
            return -np.pi/2+self.phase0
        return 0

@dataclass
class RectFunction:
    period: float
    amplitude: float
    phase: float = 0
    offset: float = 0
    wavetype: str = 'rect'

    @property
    def time_function(self) -> TimeDomainFunction:
        t0 = self.phase/2/np.pi*self.period
        return np.vectorize(lambda t: self.amplitude + self.offset if (t+t0) % self.period < self.period/2 else -self.amplitude + self.offset)

class RectFunctionHarmonics(AbstractHarmonicCoefficients):
    def _amplitude_coefficient(self, n: int) -> float:
        if n == 0:
            return self.offset0
        if n%2 == 0:
            return 0
        return 4/n/np.pi*self.amplitude0

    def _phase_coefficient(self, n: int) -> float:
        if n%2 == 0:
            return 0
        return -np.pi/2+n*self.phase0

@dataclass
class TriFunction:
    period: float
    amplitude: float
    phase: float
    offset: float = 0
    wavetype: str = 'tri'

    @property
    def time_function(self) -> TimeDomainFunction:
        t0 = self.phase/2/np.pi*self.period
        mod = lambda t: np.mod(t, self.period)
        return np.vectorize(lambda t: self.amplitude*(1-4/self.period*mod(t+t0))+self.offset if mod(t+t0) < self.period/2 else self.amplitude*(-3+4/self.period*(mod(t+t0)))+self.offset)

class TriFunctionHarmonics(AbstractHarmonicCoefficients):
    def _amplitude_coefficient(self, n: int) -> float:
        if n == 0:
            return self.offset0
        if n%2 == 0:
            return 0
        return 8/n/n/np.pi/np.pi*self.amplitude0

    def _phase_coefficient(self, n: int) -> float:
        if n%2 == 0:
            return 0
        return n*self.phase0

@dataclass
class SawFunction:
    period: float
    amplitude: float
    phase: float
    offset: float = 0
    wavetype: str = 'saw'

    @property
    def time_function(self) -> TimeDomainFunction:
        t0 = self.phase/2/np.pi*self.period
        mod = lambda t: np.mod(t, self.period)
        return np.vectorize(lambda t: self.amplitude*(2/self.period*mod(t+t0)-1)+self.offset)

class SawFunctionHarmonics(AbstractHarmonicCoefficients):
    def _amplitude_coefficient(self, n: int) -> float:
        if n == 0:
            return self.offset0
        return -2/n/np.pi*self.amplitude0

    def _phase_coefficient(self, n: int) -> float:
        if n == 0:
            return 0
        return -np.pi/2+n*self.phase0

fourier_series_mapping: dict[Type[PeriodicFunction], Type[HarmonicCoefficients]] = {
    ConstantFunction: ConstFunctionHarmonics,
    CosFunction: CosFunctionHarmonics,
    SinFunction: SinFunctionHarmonics,
    RectFunction: RectFunctionHarmonics,
    TriFunction: TriFunctionHarmonics,
    SawFunction: SawFunctionHarmonics,
}

periodic_functions : PeriodicFunctionList = list(fourier_series_mapping.keys())

def fourier_series(time_function: PeriodicFunction) -> HarmonicCoefficients:
    try:
        return fourier_series_mapping[type(time_function)](amplitude0=time_function.amplitude, phase0=time_function.phase, offset0=time_function.offset)
    except KeyError:
        raise TransformationError(f'No fourier coefficents found for time function of type {type(time_function).__name__}')

def periodic_function(wavetype: str) -> Type[PeriodicFunction]:
    try:
        return [pf for pf in periodic_functions if pf.wavetype == wavetype][0]
    except IndexError:
        raise UnknownWavetype(f'Periodic function of type {wavetype} is unknown.')


from schemdraw.elements import STYLE_IEC
from schemdraw.elements import STYLE_IEEE
from .DIN.style import STYLE_DIN
from schemdraw.elements import style as schemdraw_style

circuit_styles = {
    'IEC' : STYLE_IEC,
    'IEEE' : STYLE_IEEE,
    'DIN' : STYLE_DIN
}

def select(style: str = '') -> None:
    schemdraw_style(style=circuit_styles.get(style, STYLE_IEEE))



from dataclasses import dataclass
from . import node_analysis as na
from .solution import NodalAnalysisSolution
from ..network import Network
from ..solution import NetworkSolution
from .. import matrix_operations as mo

@dataclass
class NodalAnalysisBiasPointSolution(NodalAnalysisSolution):
    matrix_ops: mo.MatrixOperations = mo.NumPyMatrixOperations()

    def __post_init__(self) -> None:
        A = na.nodal_analysis_coefficient_matrix(self.network, matrix_ops=self.matrix_ops, node_mapper=self.node_mapper, source_mapper=self.voltage_source_mapper)
        b = na.nodal_analysis_constants_vector(self.network, matrix_ops=self.matrix_ops, node_mapper=self.node_mapper, current_source_mapper=self.current_source_mapper, voltage_source_mapper=self.voltage_source_mapper)
        try:
            self._solution_vector = self.matrix_ops.solve(A, b)
        except self.matrix_ops.matrix_inversion_exception:
            self._solution_vector = (float('nan'),) * len(b)

    def get_potential(self, node_id: str) -> complex:
        if node_id == self.network.node_zero_label:
            return 0
        return self._potentials[self._node_mapping[node_id]]

    def get_current(self, branch_id: str) -> complex:
        if branch_id in self._voltage_source_mapping.keys:
            return self._voltage_source_currents[self._voltage_source_mapping[branch_id]]
        if self.network[branch_id].element.is_ideal_current_source:
            return complex(self.network[branch_id].element.I)
        if self.network[branch_id].element.is_current_source:
            return - (self.network[branch_id].element.I + self.get_voltage(branch_id)/self.network[branch_id].element.Z)
        branch = self.network[branch_id]
        return self.get_voltage(branch_id)/branch.element.Z

def open_circuit_voltage(network: Network, node1: str, node2: str) -> complex:
    solution = NodalAnalysisBiasPointSolution(network)
    if node1 == node2:
        return 0
    phi1 = solution.get_potential(node_id=node1)
    phi2 = solution.get_potential(node_id=node2)
    return phi1-phi2

def short_circuit_current(network: Network, node1: str, node2: str) -> complex:
    Z = na.open_circuit_impedance(network, node1, node2)
    V = open_circuit_voltage(network, node1, node2)
    return V/Z

def nodal_analysis_bias_point_solver(network: Network) -> NetworkSolution:
    return NodalAnalysisBiasPointSolution(network)

def symbolic_nodal_analysis_bias_point_solver(network: Network) -> NetworkSolution:
    return NodalAnalysisBiasPointSolution(network, matrix_ops=mo.SymPyMatrixOperations())


from .norten_thevenin_elements import NortenTheveninElement, NortenElement, TheveninElement

def impedance(name : str, Z : complex) -> NortenTheveninElement:
    return NortenElement(Z=Z, V=0, name=name, type='impedance')

def admittance(name : str, Y : complex) -> NortenTheveninElement:
    return TheveninElement(Y=Y, I=0, name=name, type='admittance')

def resistor(name : str, R : float) -> NortenTheveninElement:
    return NortenElement(Z=R, V=0, name=name, type='resistor')

def conductor(name : str, G : float) -> NortenTheveninElement:
    return TheveninElement(Y=G, I=0, name=name, type='conductor')

def load(name : str, P : float, V_ref: float = -1, I_ref: float = -1, Q : float = 0) -> NortenTheveninElement:
    if V_ref < 0 and I_ref < 0:
        raise AttributeError('A reference voltage or reference current has to be defined for a load.')
    if V_ref > 0 and I_ref > 0:
        raise AttributeError('You can only define a reference voltage or a reference current for a load.')
    if V_ref == 0 and I_ref < 0:
        raise ValueError('Reference voltage must be greater than zero.')
    if I_ref <= 0 and V_ref < 0:
        raise ValueError('Reference current must be greater than zero.')
    if V_ref > 0:
        return TheveninElement(Y=complex(P, Q)/V_ref**2, I=0, name=name, type='load')
    return NortenElement(Z=complex(P, Q)/I_ref**2, V=0, name=name, type='load')

def voltage_source(name : str, V : complex, Z : complex = 0) -> NortenTheveninElement:
    return NortenElement(V=V, Z=Z, name=name, type='voltage_source')

def current_source(name : str, I : complex, Y : complex = 0) -> NortenTheveninElement:
    return TheveninElement(I=I, Y=Y, name=name, type='current_source')

def open_circuit(name : str) -> NortenTheveninElement:
    return TheveninElement(I=0, Y=0, name=name, type='open_circuit')

def short_circuit(name : str) -> NortenTheveninElement:
    return NortenElement(V=0, Z=0, name=name, type='short_circuit')



from pathlib import Path
import json
import yaml
import yaml.parser, yaml.scanner
import re
import numpy as np
from typing import Callable, TypeVar

T = TypeVar('T')

# Define a custom constructor for scientific notation
def scientific_notation_constructor(loader, node):
    value = loader.construct_scalar(node)
    try:
        return float(value)
    except ValueError:
        return value

# Add the custom constructor to the SafeLoader
yaml.SafeLoader.add_implicit_resolver(
    'tag:yaml.org,2002:float',
    re.compile(r'''^(?:[-+]?(?:[0-9][0-9_]*(?:\.[0-9_]*)?|\.[0-9_]+)(?:[eE][-+]?[0-9]+)?)$'''),
    list('-+0123456789.')
)
yaml.SafeLoader.add_constructor('tag:yaml.org,2002:float', scientific_notation_constructor)

deserializers = {
    'json': json.loads,
    'yaml': yaml.safe_load,
    'yml': yaml.safe_load
}

serializers = {
    'json': json.dumps,
    'yaml': yaml.dump,
    'yml': yaml.dump
}

class FormatError(Exception):
    def __init__(self, format: str) -> None:
        super().__init__(f'Cannot parse data as {format}.')
        self.format = format

format_errors : tuple[type[Exception], ...] = (
    json.JSONDecodeError,
    yaml.parser.ParserError,
    yaml.scanner.ScannerError
)

def dictify_complex_values(data: dict) -> dict:
    for key, value in data.items():
        if isinstance(value, complex):
            data[key] = {'real': value.real, 'imag': value.imag}
    return data

def undictify_complex_values(data: dict) -> dict:
    for key, value in data.items():
        if isinstance(value, dict) and sorted(list(value.keys())) == sorted(['real', 'imag']):
            data[key] = complex(value['real'], value['imag'])
        if isinstance(value, dict) and sorted(list(value.keys())) == sorted(['abs', 'phase']):
            if value['abs'] < 0:
                raise ValueError("abs value of '{key}' may not be negative")
            data[key] = value['abs'] * complex(np.cos(value['phase']), np.sin(value['phase']))
        if isinstance(value, dict) and sorted(list(value.keys())) == sorted(['abs', 'phase_deg']):
            if value['abs'] < 0:
                raise ValueError("abs value of '{key}' may not be negative")
            phase_rad = np.deg2rad(value['phase_deg'])
            data[key] = value['abs'] * complex(np.cos(phase_rad), np.sin(phase_rad))
    return data

def dictify_all_complex_values(data: dict) -> dict:
    for key, value in data.items():
        if isinstance(value, dict):
            data[key] = dictify_all_complex_values(value)
    return data

def undictify_all_complex_values(data: dict) -> dict:
    for key, value in data.items():
        if isinstance(value, dict):
            data[key] = undictify_all_complex_values(value)
        if isinstance(value, list):
            data[key] = [undictify_all_complex_values(v) for v in value]
    return undictify_complex_values(data)

def serialize(data: T, format: str, dict_processor: Callable[[T], dict] = dictify_all_complex_values) -> str:
    serializer = serializers.get(format, None)
    if serializer is None:
        raise ValueError('Unknown data format {format}.')
    return serializer(dict_processor(data))

def dump(file: str, data: T, dump_fcn: Callable[[T, str], str] = serialize) -> None:
    file_name = Path(file)
    suffix = file_name.suffix[1:]
    with open(file_name, 'w') as f:
        f.write(dump_fcn(data, suffix))

def deserialize(data: str, format: str, dict_preprocessor: Callable[[dict], T] = undictify_all_complex_values) -> T:
    deserializer = deserializers.get(format, None)
    if deserializer is None:
        raise ValueError('Unknown data format {format}.')
    try:
        return dict_preprocessor(deserializer(data))
    except format_errors as e:
        raise FormatError(format) from e

def load(file: str, deserialize_fcn: Callable[[str, str], T] = deserialize) -> T:
    file_name = Path(file)
    suffix = file_name.suffix[1:]
    with open(file_name) as f:
        return deserialize_fcn(f.read(), suffix)



import numpy as np
from matplotlib.patches import Polygon
from matplotlib.axes import Axes

def get_aspect(ax: Axes) -> float:
    figW, figH = ax.get_figure().get_size_inches()
    _, _, w, h = ax.get_position().bounds
    disp_ratio = (figH * h) / (figW * w)
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    data_ratio = (ylim[1]-ylim[0]) / (xlim[1]-xlim[0])
    return disp_ratio / data_ratio

def arrow_head(ax: Axes, xy0, height, width, angle, color='black', **kwargs):
    def A(phi: float) -> np.matrix:
        xy_ratio = get_aspect(ax)
        return np.matrix([[np.cos(phi), -np.sin(phi)*xy_ratio], [np.sin(phi)/xy_ratio, np.cos(phi)]])
    def triangular_shape(abs_height: float, abs_width: float) -> np.matrix:
        return np.matrix([np.array([-abs_height, 0, -abs_height]),
                          np.array([abs_width/2, 0, -abs_width/2])])
    def offset(x: float, y: float) -> np.matrix:
        return np.matrix([[x], [y]])
    def create_patch(polygon, color, **kwargs):
        if 'linewidth' in kwargs.keys():
            kwargs.pop('linewidth')
        return Polygon(polygon.T, color=color, linewidth=0, zorder=100, **kwargs)
    triangular = A(angle)*triangular_shape(width, height)+offset(xy0[0], xy0[1])
    patch = create_patch(triangular, color, **kwargs)
    ax.add_patch(patch)

def arrow(ax: Axes, xy0, xy1, height, width, tail=False, color='black', **kwargs) -> None:
    dx = xy1[0]-xy0[0]
    dy = xy1[1]-xy0[1]
    alpha_r = np.arctan2(dy*get_aspect(ax), dx)
    alpha = np.arctan2(dy, dx)
    arrow_head(
        ax=ax,
        xy0=xy1,
        height=height,
        width=width,
        angle=alpha_r,
        color=color,
        **kwargs)
    if tail:
        arrow_head(
            ax=ax,
            xy0=xy0,
            height=height,
            width=width,
            angle=alpha-np.pi,
            color=color,
            **kwargs)
    if get_aspect(ax) > 1:
        aspect = get_aspect(ax)
    else:
        aspect = 1
    if tail:
        xy0[0]+=height*np.cos(alpha)/aspect
        xy0[1]+=height*np.sin(alpha)/aspect
    xy1[0]-=height*np.cos(alpha)/aspect
    xy1[1]-=height*np.sin(alpha)/aspect
    ax.plot([xy0[0], xy1[0]], [xy0[1], xy1[1]], color=color, **kwargs)

def complex_pointer(ax: Axes, z0, z1, width=0.05, height=0.05, **kwargs) -> None:
    xlim, ylim = ax.get_xlim(), ax.get_ylim()
    delta_x, delta_y = xlim[1]-xlim[0], ylim[1]-ylim[0]
    max_stretch = max(delta_x, delta_y)
    width *= max_stretch
    height *= max_stretch
    if z0 != z1:
        arrow(ax, [np.real(z0), np.imag(z0)], [np.real(z1), np.imag(z1)], width, height, **kwargs)



from typing import Any, Callable
from .circuit import Circuit, Component
from .Components import components as cp
from .Components import symbolic_components as s_cp
from dataclasses import asdict
from .. import dump_load
import functools

class UnidentifiedComponent(Exception):
    ...

class UnknownCircuitComponent(Exception):
    ...

class IncorrectComponentInformation(Exception):
    ...

def numeric_component_factory(*_, factory_fcn: tuple[Callable[..., cp.Component], Callable[..., cp.Component]], numeric_keys: tuple[str, ...], **kwargs) -> cp.Component:
    def is_numeric(value: str) -> bool:
        try:
            float(value)
            return True
        except ValueError:
            return False
    numeric_values = [kwargs.get(key, '') for key in numeric_keys]
    if all(is_numeric(value) for value in numeric_values):
        return factory_fcn[0](**kwargs)
    return factory_fcn[1](**kwargs)

circuit_component_translators : dict[str, Callable[..., cp.Component]] = {
    "resistor" : functools.partial(numeric_component_factory, factory_fcn=(cp.resistor, s_cp.resistor), numeric_keys=('R',)),
    "conductance" : functools.partial(numeric_component_factory, factory_fcn=(cp.conductance, s_cp.conductance), numeric_keys=('G',)),
    "impedance" : functools.partial(numeric_component_factory, factory_fcn=(cp.impedance, s_cp.impedance), numeric_keys=('Z',)),
    "admittance" : functools.partial(numeric_component_factory, factory_fcn=(cp.admittance, s_cp.admittance), numeric_keys=('Y',)),
    "dc_voltage_source" : functools.partial(numeric_component_factory, factory_fcn=(cp.dc_voltage_source, s_cp.voltage_source), numeric_keys=('V',)),
    "ac_voltage_source" : functools.partial(numeric_component_factory, factory_fcn=(cp.ac_voltage_source, s_cp.voltage_source), numeric_keys=('V', 'f')),
    "complex_voltage_source" : functools.partial(numeric_component_factory, factory_fcn=(cp.complex_voltage_source, s_cp.voltage_source), numeric_keys=('V',)),
    "dc_current_source" : functools.partial(numeric_component_factory, factory_fcn=(cp.dc_current_source, s_cp.current_source), numeric_keys=('I',)),
    "ac_current_source" : functools.partial(numeric_component_factory, factory_fcn=(cp.ac_current_source, s_cp.current_source), numeric_keys=('I', 'f')),
    "complex_current_source" : functools.partial(numeric_component_factory, factory_fcn=(cp.complex_current_source, s_cp.current_source), numeric_keys=('I',)),
}

def generate_component(component: dict[str, Any]) -> Component:
    component = component.copy()
    try:
        component_id = component['id']
    except KeyError:
        raise UnidentifiedComponent(f'Unidentified component')
    try:
        component_value = component.pop('value', {})
    except KeyError:
        raise IncorrectComponentInformation(f"Missing value of component '{component_id}'.")
    try:
        component_type = component.pop('type')
    except KeyError:
        raise IncorrectComponentInformation(f"Missing type information of component '{component_id}'.")
    try:
        component_nodes = component['nodes']
    except KeyError:
        raise IncorrectComponentInformation(f"Missing node information of component '{component_id}'.")
    try:
        component_factory = circuit_component_translators[component_type]
    except KeyError:
        raise UnknownCircuitComponent(f"Unknown type '{component_type}' of component '{component_id}' is unknown.")
    return component_factory(id=component_id, nodes=component_nodes, **component_value)

def undictify_circuit(circuit: dict) -> Circuit:
    return Circuit([generate_component(entry) for entry in circuit['components']])

deserialize = functools.partial(dump_load.deserialize, dict_preprocessor=undictify_circuit)
load = functools.partial(dump_load.load, deserialize_fcn=deserialize)

def dictify_circuit(circuit: Circuit) -> dict:
    return {'components' : [asdict(c) for c in circuit.components]}

serialize = functools.partial(dump_load.serialize, dict_processor=dictify_circuit)
save = functools.partial(dump_load.dump, dump_fcn=serialize)



import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes as Axis

from typing import Callable

class LayoutError(Exception):
    ...

color = {
    'black' : (0, 0, 0),
    'grey' : (.7, .7, .7),
    'lightgrey' : (.9, .9, .9),
    'darkgrey' : (.5, .5, .5),
    'darkdarkgrey' : (.2, .2, .2),
    'darkblue' : (.13, .22, .39),
    'lightblue' : (.4, .7, 1),
    'red' : (0.721, 0, 0),
    'darkred' : (0.8, 0, 0),
    'green' : (0, 0.45, 0.333),
    'blue' : (.18, .33, .59),
    'purple' : (.44, .19, .63),
    'white' : (1, 1, 1)
}

Axes = tuple[Axis, ...]
FigureAxes = tuple[Figure, Axes]
Layout = Callable[[], FigureAxes]
PlotFcn = Callable[[Figure, Axes], FigureAxes]

def figure_default() -> FigureAxes:
    fig, ax = plt.subplots()
    return fig, (ax,)
    
def figure_two_rows() -> FigureAxes:
    fig, ax = plt.subplots(nrows=2, sharex=True)
    return fig, tuple(ax)

def apply_plt_fcn(fig, ax, *args: PlotFcn) -> FigureAxes:
    for plt_fcn in args:
        plt_fcn(fig, ax)
    return fig, ax

def legend(yoffset:float=1.1) -> Callable[[PlotFcn], PlotFcn]:
    def decorator(plot_fcn: PlotFcn) -> PlotFcn:
        def wrapper(fig: Figure, ax: Axes, *args, **kwargs) -> FigureAxes:
            fig, ax = plot_fcn(fig, ax, *args, **kwargs)
            ax[0].legend(
                handles=[line for line in ax[0].lines if not line._label.startswith('_')],
                ncol=len(ax[0].lines),
                loc='upper center',
                bbox_to_anchor=(0.5, yoffset),
                frameon=False
            )
            return fig, ax
        return wrapper
    return decorator

def grid(visible:bool=True) -> Callable[[PlotFcn], PlotFcn]:
    def decorator(plot_fcn: PlotFcn) -> PlotFcn:
        def wrapper(fig: Figure, ax: Axes, *args, **kwargs) -> FigureAxes:
            for a in ax:
                a.grid(visible=visible, zorder=-1)
            return plot_fcn(fig, ax, *args, **kwargs)
        return wrapper
    return decorator

def equal_scaling() -> Callable[[PlotFcn], PlotFcn]:
    def decorator(plot_fcn: PlotFcn) -> PlotFcn:
        def wrapper(fig: Figure, ax: Axes, *args, **kwargs) -> FigureAxes:
            for a in ax:
                a.set_aspect('equal', 'box')
            return plot_fcn(fig, ax, *args, **kwargs)
        return wrapper
    return decorator

def xlim_bottom(xmin:float=0, xmax:float=-1) -> Callable[[PlotFcn], PlotFcn]:
    def decorator(plot_fcn: PlotFcn) -> PlotFcn:
        def wrapper(fig: Figure, ax: Axes, *args, **kwargs) -> FigureAxes:
            if xmax >= xmin:
                ax[-1].set_xlim(left=xmin, right=xmax)
            return plot_fcn(fig, ax, *args, **kwargs)
        return wrapper
    return decorator

def xlabel_bottom(label:str='') -> Callable[[PlotFcn], PlotFcn]:
    def decorator(plot_fcn: PlotFcn) -> PlotFcn:
        def wrapper(fig: Figure, ax: Axes, *args, **kwargs) -> FigureAxes:
            ax[-1].set_xlabel(label)
            return plot_fcn(fig, ax, *args, **kwargs)
        return wrapper
    return decorator

def nyquist_like_plot(ax_lim:tuple[float, float, float, float], xlabel:str='', ylabel:str='') -> Callable[[PlotFcn], PlotFcn]:
    def decorator(plot_fcn: PlotFcn) -> PlotFcn:
        @equal_scaling()
        def wrapper(fig: Figure, ax: Axes, *args, **kwargs) -> FigureAxes:
            ax[0].set_xlabel(xlabel)
            ax[0].set_ylabel(ylabel)
            ax[0].set_xlim(left=ax_lim[0], right=ax_lim[1])
            ax[0].set_ylim(bottom=ax_lim[2], top=ax_lim[3])
            return plot_fcn(fig, ax, *args, **kwargs)
        return wrapper
    return decorator

def frequencydomain_plot(ylabel:str='') -> Callable[[PlotFcn], PlotFcn]:
    def decorator(plot_fcn: PlotFcn) -> PlotFcn:
        def wrapper(fig: Figure, ax: Axes, *args, **kwargs) -> FigureAxes:
            ax[0].set_ylabel(ylabel)
            ax[1].set_ylabel(ylabel)
            return plot_fcn(fig, ax, *args, **kwargs)
        return wrapper
    return decorator

def bode_like_plot(wmin:float=0, wmax:float=-1, logw:bool=False, logy:bool=False, xlabel:str='ω→', ylabel:str='') -> Callable[[PlotFcn], PlotFcn]:
    def decorator(plot_fcn: PlotFcn) -> PlotFcn:
        @xlabel_bottom(xlabel)
        @xlim_bottom(xmin=wmin, xmax=wmax)
        def wrapper(fig: Figure, ax: Axes, *args, **kwargs) -> FigureAxes:
            if len(ylabel) > 0:
                ax[-2].set_ylabel(f'|{ylabel}|→')
                ax[-1].set_ylabel(f'arg{{{ylabel}}}→')
            ax[-2].set_yscale('log' if logy else 'linear')
            ax[-1].set_xscale('log' if logw else 'linear')
            return plot_fcn(fig, ax, *args, **kwargs)
        return wrapper
    return decorator


from ..Circuit.solution import CircuitSolution, DCSolution, ComplexSolution
from ..Circuit.circuit import Circuit

from . import Elements as elm
from . import Display as dsp
from .DiagramTranslator import SchematicDiagramParser, circuit_translator

from dataclasses import dataclass, field
from typing import Protocol
import numpy as np

class DiagramSolution(Protocol):
    def get_voltage(self, name: str, reverse: bool) -> str:
        ...
    def get_current(self, name: str, reverse: bool) -> str:
        ...
    def get_power(self, name: str) -> str:
        ...
    def get_potential(self, name: str) -> str:
        ...
    @property
    def solution(self) -> CircuitSolution:
        ...

@dataclass
class EmptyDiagramSolution:
    solution: CircuitSolution = field(default_factory=lambda: DCSolution(circuit=Circuit([])))
    def get_voltage(self, name: str, reverse: bool) -> str:
        return f'V[{name}]'
    def get_current(self, name: str, reverse: bool) -> str:
        return f'I[{name}]'
    def get_power(self, name: str) -> str:
        return f'P[{name}]'
    def get_potential(self, name: str) -> str:
        return f'φ[{name}]'

@dataclass
class TimeDomainSteadyStateDiagramSolution:
    solution: ComplexSolution
    deg: bool = False
    hertz: bool = False
    sin: bool = False
    precision: int = 3

    def get_voltage(self, name: str, reverse: bool) -> str:
        sign = -1 if reverse else 1
        v = self.solution.get_voltage(name)
        if np.isnan(v):
            return f'v[{name}](t)'
        return dsp.print_sinosoidal(
            value=sign*v,
            unit='V',
            precision=self.precision,
            w=self.solution.w,
            sin=self.sin,
            deg=self.deg,
            hertz=self.hertz
        )

    def get_current(self, name: str, reverse: bool) -> str:
        sign = -1 if reverse else 1
        i = self.solution.get_voltage(name)
        if np.isnan(i):
            return f'i[{name}](t)'
        return dsp.print_sinosoidal(
            value=sign*i,
            unit='A',
            precision=self.precision,
            w=self.solution.w,
            sin=self.sin,
            deg=self.deg,
            hertz=self.hertz
        )

    def get_power(self, name: str) -> str:
        p = self.solution.get_voltage(name)
        if np.isnan(p):
            return f'p[{name}](t)'
        return dsp.print_sinosoidal(
            value=self.solution.get_power(name),
            unit='W',
            precision=self.precision,
            w=self.solution.w,
            sin=self.sin,
            deg=self.deg,
            hertz=self.hertz
        )

    def get_potential(self, name: str) -> str:
        phi = self.solution.get_voltage(name)
        if np.isnan(phi):
            return f'φ[{name}](t)'
        return dsp.print_sinosoidal(
            value=self.solution.get_potential(name),
            unit='V',
            precision=self.precision,
            w=self.solution.w,
            sin=self.sin,
            deg=self.deg,
            hertz=self.hertz
        )

@dataclass
class ComplexNetworkDiagramSolution:
    solution: ComplexSolution
    deg: bool = False
    polar: bool = False
    precision: int = 3

    def get_voltage(self, name: str, reverse: bool) -> str:
        sign = -1 if reverse else 1
        v = self.solution.get_voltage(name)
        if np.isnan(v):
            return f'V({name})'
        return dsp.print_complex(
            value=sign*v,
            unit='V',
            precision=self.precision,
            polar=self.polar,
            deg=self.deg
        )

    def get_current(self, name: str, reverse: bool) -> str:
        sign = -1 if reverse else 1
        i = self.solution.get_current(name)
        if np.isnan(i):
            return f'I({name})'
        return dsp.print_complex(
            value=sign*i,
            unit='A',
            precision=self.precision,
            polar=self.polar,
            deg=self.deg
        )

    def get_power(self, name: str) -> str:
        s = self.solution.get_power(name)
        if np.isnan(s):
            return f'S({name})'
        return dsp.print_complex(
            value=s,
            unit='VA',
            precision=self.precision,
            polar=self.polar,
            deg=self.deg
        )

    def get_potential(self, name: str) -> str:
        phi = self.solution.get_potential(name)
        if np.isnan(phi):
            return f'φ({name})'
        return dsp.print_complex(
            value=phi,
            unit='V',
            precision=self.precision,
            polar=self.polar,
            deg=self.deg
        )

@dataclass
class RealNetworkDiagramSolution:
    solution: DCSolution
    precision: int = 3

    def get_voltage(self, name: str, reverse: bool) -> str:
        sign = -1 if reverse else 1
        v = self.solution.get_voltage(name)
        if np.isnan(v):
            return f'V({name})'
        return dsp.print_real(sign*v, unit='V', precision=self.precision)

    def get_current(self, name: str, reverse: bool) -> str:
        sign = -1 if reverse else 1
        i = self.solution.get_current(name)
        if np.isnan(i):
            return f'I({name})'
        return dsp.print_real(sign*i, unit='A', precision=self.precision)

    def get_power(self, name: str) -> str:
        p = self.solution.get_power(name)
        if np.isnan(p):
            return f'P({name})'
        return dsp.print_active_power(p, precision=self.precision)

    def get_potential(self, name: str) -> str:
        phi = self.solution.get_potential(name)
        if np.isnan(phi):
            return f'φ({name})'
        return dsp.print_real(phi, unit='V', precision=self.precision)

@dataclass
class SchematicDiagramSolution:
    diagram_parser: SchematicDiagramParser
    solution: DiagramSolution

    def draw_voltage(self, name: str, reverse: bool = False, offset: float = 0) -> elm.VoltageLabel:
        element = self.diagram_parser.get_element(name)
        vlabel = self.solution.get_voltage(name=name, reverse=reverse)
        return elm.VoltageLabel(element, vlabel=vlabel, reverse=reverse if not element.is_reverse else not reverse, color=dsp.blue, ofst=offset)

    def draw_current(self, name: str, reverse: bool = False, end: bool = False) -> elm.CurrentLabel:
        element = self.diagram_parser.get_element(name)
        ilabel = self.solution.get_current(name=name, reverse=reverse if not end else not reverse)
        return elm.CurrentLabel(element, ilabel=ilabel, reverse=reverse if not element.is_reverse else not reverse, start=not end, color=dsp.red)

    def draw_power(self, name: str, offset: float = 0) -> elm.PowerLabel:
        element = self.diagram_parser.get_element(name)
        plabel = self.solution.get_power(name=name)
        return elm.PowerLabel(element, plabel=plabel, color=dsp.green, offset=offset)

    def draw_potential(self, name: str, loc:str = '') -> elm.LabelNode:
        element = self.diagram_parser.get_element(name)
        phi_label = self.solution.get_potential(name=name)
        return elm.LabelNode(id_loc=loc, name=phi_label, at=element.absdrop[0], color=dsp.blue)

def empty_solution(schematic: elm.Schematic) -> SchematicDiagramSolution:
    return SchematicDiagramSolution(
        diagram_parser=SchematicDiagramParser(schematic),
        solution=EmptyDiagramSolution()
    )

def single_frequency_time_domain_steady_state_solution(schematic: elm.Schematic, w: float = 0, sin: bool = False, deg: bool = False, hertz: bool = False) -> SchematicDiagramSolution:
    digagram_parser = SchematicDiagramParser(schematic)
    solution = TimeDomainSteadyStateDiagramSolution(
        solution=ComplexSolution(circuit=circuit_translator(schematic), w=w),
        deg=deg,
        hertz=hertz,
        sin=sin
    )
    return SchematicDiagramSolution(
        diagram_parser=digagram_parser,
        solution=solution
    )

def single_frequency_complex_solution(schematic: elm.Schematic, w: float = 0, precision: int = 3, polar: bool = False, deg: bool = False) -> SchematicDiagramSolution:
    diagram_parser = SchematicDiagramParser(schematic)
    solution = ComplexNetworkDiagramSolution(
        solution=ComplexSolution(circuit=circuit_translator(schematic), w=w),
        deg=deg,
        polar=polar,
        precision=precision
    )
    return SchematicDiagramSolution(
        diagram_parser=diagram_parser,
        solution=solution,
    )

def complex_solution(schematic: elm.Schematic, precision: int = 3, polar: bool = False, deg: bool = False) -> SchematicDiagramSolution:
    diagram_parser = SchematicDiagramParser(schematic)
    solution = ComplexNetworkDiagramSolution(
        solution=ComplexSolution(circuit=circuit_translator(schematic)),
        deg=deg,
        polar=polar,
        precision=precision
    )
    return SchematicDiagramSolution(
        diagram_parser=diagram_parser,
        solution=solution,
    )

def real_solution(schematic: elm.Schematic, precision: int = 3) -> SchematicDiagramSolution:
    diagram_parser = SchematicDiagramParser(schematic)
    solution = RealNetworkDiagramSolution(
        solution=DCSolution(circuit=circuit_translator(schematic)),
        precision=precision
    )
    return SchematicDiagramSolution(
        diagram_parser=diagram_parser,
        solution=solution
    )


from .network import Network, Branch
from .elements import NortenTheveninElement, impedance, admittance

def switch_ground_node(network: Network, new_ground: str) -> Network:
    return Network(network.branches, new_ground)

def remove_element(network: Network, element: str) -> Network:
    branch_to_remove = network[element]
    if branch_to_remove.node1 == network.node_zero_label and len(network.branches_connected_to(branch_to_remove.node1)) == 1:
       network = rename_node(network, branch_to_remove.node2, network.node_zero_label) 
    if branch_to_remove.node2 == network.node_zero_label and len(network.branches_connected_to(branch_to_remove.node2)) == 1:
       network = rename_node(network, branch_to_remove.node1, network.node_zero_label) 
    branches = [b for b in network.branches if b.element.name != element]
    return Network(branches, node_zero_label=network.node_zero_label)
    
def remove_elements(network: Network, elements: list[str]) -> Network:
    for element in elements:
        network = remove_element(network, element)
    return network

def rename_node(network: Network, old_node: str, new_node: str) -> Network:
    def rename(branch: Branch) -> Branch:
        if branch.node1 == old_node:
            branch = Branch(new_node, branch.node2, branch.element)
        if branch.node2 == old_node:
            branch = Branch(branch.node1, new_node, branch.element)
        return branch
    branches = [rename(b) for b in network.branches]
    return Network(branches, node_zero_label=network.node_zero_label)

def remove_open_circuit_elements(network: Network) -> Network:
    return Network([b for b in network.branches if not b.element.is_open_circuit], node_zero_label=network.node_zero_label)

def remove_short_circuit_elements(network: Network, keep: list[NortenTheveninElement] = []) -> Network:
    branches = network.branches
    short_circuits = [b for b in network.branches if b.element.is_short_circuit and b.element not in keep]
    short_circuit_nodes = [(vs.node1, vs.node2) if not network.is_zero_node(vs.node1) else (vs.node2, vs.node1) for vs in short_circuits]
    for an, rn in short_circuit_nodes:
        branches = [Branch(rn, b.node2, b.element) if b.node1 == an else b for b in branches]
        branches = [Branch(b.node1, rn, b.element) if b.node2 == an else b for b in branches]
        branches = [b for b in branches if b.node1 != b.node2]
    return Network(branches, node_zero_label=network.node_zero_label)

def short_circuitify_voltage_sources(network: Network, keep: list[NortenTheveninElement] = []) -> Network:
    def zero_in_voltage(branch: Branch) -> Branch:
        return Branch(branch.node1, branch.node2, impedance(branch.element.name, branch.element.Z))
    def is_intended_voltage_source(branch: Branch) -> bool:
        if branch.element in keep:
            return False
        if branch.element.is_voltage_source:
            return True
        return False
    return Network(
        branches=[zero_in_voltage(b) if is_intended_voltage_source(b) else b for b in network.branches],
        node_zero_label=network.node_zero_label
    )

def open_circuitify_current_sources(network: Network, keep: list[NortenTheveninElement] = []) -> Network:
    def zero_in_current(branch: Branch) -> Branch:
        return Branch(branch.node1, branch.node2, admittance(branch.element.name, branch.element.Y))
    def is_intended_current_source(branch: Branch) -> bool:
        if branch.element in keep:
            return False
        if branch.element.is_current_source:
            return True
        return False
    return Network(
        branches=[zero_in_current(b) if is_intended_current_source(b) else b for b in network.branches],
        node_zero_label=network.node_zero_label
    )

def remove_ideal_current_sources(network: Network, keep: list[NortenTheveninElement] = []) -> Network:
    return remove_open_circuit_elements(open_circuitify_current_sources(network, keep=keep))

def remove_ideal_voltage_sources(network: Network, keep: list[NortenTheveninElement] = []) -> Network:
    return remove_short_circuit_elements(short_circuitify_voltage_sources(network, keep=keep), keep=keep)

def passive_network(network: Network, keep: list[NortenTheveninElement] = []) -> Network:
    return remove_ideal_voltage_sources(remove_ideal_current_sources(network, keep=keep), keep=keep)


from typing import Optional
from matplotlib.axes import Axes
from typing import Callable
import CircuitCalculator.SimpleCircuit.Elements as elm
import CircuitCalculator.SimpleCircuit.DiagramParser as dp
import CircuitCalculator.SimpleCircuit.DiagramSolution as ds
import CircuitCalculator.SimpleCircuit.LampLighter as ll
from dataclasses import dataclass
from inspect import signature
from . import errors


solutions = {
    'dc': ds.real_solution,
    'real': ds.real_solution,
    'complex': ds.single_frequency_complex_solution,
    'single_frequency_time_domain': ds.single_frequency_complex_solution
}

@dataclass
class SolutionDefinition:
    data : dict

    @property
    def diagram_solution_creator(self) -> Callable[[elm.Schematic], ds.SchematicDiagramSolution]:
        solution_type = self.data.get('type', 'unknown')
        solution_fcn = solutions.get(solution_type, ds.empty_solution)
        feasible_solution_params = signature(solution_fcn).parameters.keys()
        solution_parameters = {k: v for k, v in self.data.items() if k in feasible_solution_params}
        return lambda schematic: solution_fcn(schematic=schematic, **solution_parameters)

    @property
    def voltages(self) -> list[dict]:
        voltages = self.data.get('voltages', [])
        if voltages is None:
            return []
        return voltages

    @property
    def currents(self) -> list[dict]:
        currents = self.data.get('currents', [])
        if currents is None:
            return []
        return currents

    @property
    def potentials(self) -> list[dict]:
        potentials = self.data.get('potentials', [])
        if potentials is None:
            return []
        return potentials

    @property
    def powers(self) -> list[dict]:
        powers = self.data.get('powers', [])
        if powers is None:
            return []
        return powers

element_handlers = {
    'resistor': lambda kwargs: element_factory(elm.Resistor, **kwargs),
    'conductance': lambda kwargs: element_factory(elm.Conductance, **kwargs),
    'impedance': lambda kwargs: element_factory(elm.Impedance, **kwargs),
    'admittance': lambda kwargs: element_factory(elm.Admittance, **kwargs),
    'capacitor': lambda kwargs: element_factory(elm.Capacitor, **kwargs),
    'inductance': lambda kwargs: element_factory(elm.Inductance, **kwargs),
    'line': lambda kwargs: element_factory(elm.LabeledLine, **kwargs) if 'name' in kwargs.keys() else element_factory(elm.Line, **kwargs),
    'node': lambda kwargs: element_factory(elm.Node, **kwargs),
    'lamp': lambda kwargs: element_factory(elm.Lamp, **kwargs),
    'ground': lambda kwargs: element_factory(elm.Ground, **kwargs),
    'voltage_source': lambda kwargs: element_factory(elm.VoltageSource, **kwargs),
    'ac_voltage_source': lambda kwargs: element_factory(elm.ACVoltageSource, **kwargs),
    'complex_voltage_source': lambda kwargs: element_factory(elm.ComplexVoltageSource, **kwargs),
    'current_source': lambda kwargs: element_factory(elm.CurrentSource, **kwargs),
    'ac_current_source': lambda kwargs: element_factory(elm.ACCurrentSource, **kwargs),
    'complex_current_source': lambda kwargs: element_factory(elm.ComplexCurrentSource, **kwargs),
}

def element_factory(element: type[elm.Element], name: str = '', reverse: bool = False, **kwargs) -> elm.Element:
    try:
        return element(name=name, reverse=reverse, **kwargs)
    except TypeError as e:
        missing_argument = str(e).split(":")[-1].strip()
        provided_arguments = {}
        if name != '':
            provided_arguments = {'name': name}
        provided_arguments.update(kwargs)
        raise errors.MissingArgument(missing_argument, str(provided_arguments)) from e

def transform_to_schematic_element(element: dict) -> elm.Element:
    try:
        element_type = element['type']
    except KeyError as e:
        raise errors.MissingArgument('type', str(element)) from e
    try:
        return element_handlers[element_type](element)
    except KeyError as e:
        raise errors.UnknownCircuitElement(element_type) from e

def apply_direction_and_length(element: elm.Element, direction: str = '', length: float = 1, unit: float = 1) -> elm.Element:
    if direction == 'right':
        element.right(length*unit)
    elif direction == 'left':
        element.left(length*unit)
    elif direction == 'up':
        element.up(length*unit)
    elif direction == 'down':
        element.down(length*unit)
    return element

def apply_position(element: elm.Element, origin_element: Optional[elm.schemdraw.elements.Element] = None) -> elm.Element:
    if origin_element is None:
        return element
    return element.at(origin_element.end)

def get_placed_element(schematic: elm.Schematic, label: Optional[str] = None) -> Optional[elm.schemdraw.elements.Element]:
    if label is None:
        return None
    return schematic.elements[[se.name for se in schematic.elements].index(label)]

def fill_schematic(schematic: elm.Schematic, elements: list[elm.Element], unit: int) -> None:
    for e in elements:
        se = transform_to_schematic_element(e)
        se = apply_direction_and_length(se, e.get('direction', ''), e.get('length', 1), unit)
        se = apply_position(se, get_placed_element(schematic, e.get('place_after', None)))
        schematic += se

def fill_solution(schematic: elm.Schematic, light_lamps: bool, solution_definition: SolutionDefinition) -> None:
    if light_lamps:
        ll.light_lamps(schematic)

    try:
        solution = solution_definition.diagram_solution_creator(schematic)
    except ValueError as e:
        raise errors.IllegalElementValue(str(e)) from e
    for v in solution_definition.voltages:
        try:
            schematic += solution.draw_voltage(**v)
        except dp.UnknownElement as e:
            print(f'Cannot draw voltage of undefined element "{str(e)}".')
        except TypeError as e:
            unknown_argument = str(e).split()[-1].strip()
            raise errors.UnknownArgument(unknown_argument, 'voltages') from e
    for c in solution_definition.currents:
        try:
            schematic += solution.draw_current(**c)
        except dp.UnknownElement as e:
           print(f'Cannot draw current of undefined element "{str(e)}".')
        except TypeError as e:
            unknown_argument = str(e).split()[-1].strip()
            raise errors.UnknownArgument(unknown_argument, 'currents') from e
    for p in solution_definition.potentials:
        try:
            schematic += solution.draw_potential(**p)
        except dp.UnknownElement as e:
           print(f'Cannot draw potential of undefined node "{str(e)}".')
        except TypeError as e:
            unknown_argument = str(e).split()[-1].strip()
            raise errors.UnknownArgument(unknown_argument, 'potential') from e
    for p in solution_definition.powers:
        try:
            schematic += solution.draw_power(**p)
        except dp.UnknownElement as e:
           print(f'Cannot draw power of undefined element "{str(e)}".')
        except TypeError as e:
            unknown_argument = str(e).split()[-1].strip()
            raise errors.UnknownArgument(unknown_argument, 'powers') from e

def parse_circuit_data(data: dict) -> dict:
    circuit_definiton = data.get('circuit', {'unit': 7, 'elements': [], 'solution': {'type': None}})
    if len(circuit_definiton['elements']) == 0:
        raise errors.EmptyCircuit('No elements in circuit definition')
    return circuit_definiton

def create_schematic(circuit_data: dict) -> elm.Schematic:
    circuit_data = parse_circuit_data(circuit_data)
    unit = circuit_data.get('unit', 7)
    elements = circuit_data.get('elements', [])
    schematic = elm.Schematic(unit=unit)
    fill_schematic(schematic, elements, unit)
    return schematic

def draw_schematic(circuit_data: dict, circuit_ax: Optional[Axes] = None) -> elm.Schematic:
    circuit_data = parse_circuit_data(circuit_data)
    unit = circuit_data.get('unit', 7)
    elements = circuit_data.get('elements', [])
    light_lamps = circuit_data.get('light_lamps', False)
    solution_definition = SolutionDefinition(circuit_data.get('solution', {}))
    if circuit_ax is None:
        with elm.Schematic(unit=unit) as schematic:
            fill_schematic(schematic, elements, unit)
            fill_solution(schematic, light_lamps, solution_definition)
        return schematic
    schematic = elm.Schematic(unit=circuit_data['unit'], canvas=circuit_ax)
    fill_schematic(schematic, elements, unit)
    fill_solution(schematic, light_lamps, solution_definition)
    schematic.draw(show=False)
    return schematic